@@ -1,7 +1,7 @@
 # -*- coding: utf-8; mode: tcl; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- vim:fenc=utf-8:filetype=tcl:et:sw=4:ts=4:sts=4
 # portstartupitem.tcl
 #
-# Copyright (c) 2004-2012 The MacPorts Project
+# Copyright (c) 2004-2014, 2016-2018, 2020 The MacPorts Project
 # Copyright (c) 2006-2007 James D. Berry
 # Copyright (c) 2004,2005 Markus W. Weissman <mww@macports.org>
 # All rights reserved.
@@ -53,8 +53,15 @@
 #       We create and destroy the pidfile to track the pid we receive from the executable
 #
 #   startupitem.logfile     logpath
-#       Log to the specified file -- if not specified then output to /dev/null
-#       - for launchd, just set this as the standard out key
+#       Log stdout to the specified logfile
+#       - If not specified, then output to /dev/null
+#       - For launchd, set the stdout plist key
+#
+#   startupitem.logfile.stderr logpath
+#       Log stderr to the specified logfile
+#       - If not specified, defaults to startupitem.logfile
+#       - If cleared, disables stderr logging
+#       - For launchd, set the stderr plist key
 #
 #   startupitem.logevents   yes/no
 #       Log events to the log
@@ -64,6 +71,12 @@
 #   startupitem.autostart   yes/no
 #       Automatically load the startupitem after activating. Defaults to no.
 #
+#   startupitem.user        <user>
+#       User to run the service/daemon as.
+#
+#   startupitem.group       <group>
+#       Group to run the service/daemon as.
+#
 #   startupitem.debug       yes/no
 #       Enable additional debug logging
 #       - for launchd, sets the Debug key to true
@@ -75,32 +88,38 @@ namespace eval portstartupitem {
 }
 
 options startupitems startupitem.autostart startupitem.debug \
-        startupitem.create startupitem.executable \
+        startupitem.create startupitem.custom_file \
+        startupitem.executable startupitem.group \
         startupitem.init startupitem.install startupitem.location \
-        startupitem.logevents startupitem.logfile startupitem.name \
+        startupitem.logevents startupitem.logfile \
+        startupitem.logfile.stderr startupitem.name \
         startupitem.netchange startupitem.pidfile startupitem.plist \
         startupitem.requires startupitem.restart startupitem.start \
         startupitem.stop startupitem.type startupitem.uniquename \
-        startupitem.daemondo.verbosity
+        startupitem.user startupitem.daemondo.verbosity
 
 default startupitem.autostart   no
+default startupitem.custom_file {}
 default startupitem.debug       no
-default startupitem.executable  ""
-default startupitem.init        ""
+default startupitem.executable  {}
+default startupitem.group       {}
+default startupitem.init        {}
 default startupitem.install     {$system_options(startupitem_install)}
 default startupitem.location    LaunchDaemons
 default startupitem.logevents   no
-default startupitem.logfile     ""
+default startupitem.logfile     {}
+default startupitem.logfile.stderr {${startupitem.logfile}}
 default startupitem.name        {${subport}}
 default startupitem.netchange   no
-default startupitem.pidfile     ""
+default startupitem.pidfile     {}
 default startupitem.plist       {${startupitem.uniquename}.plist}
-default startupitem.requires    ""
-default startupitem.restart     ""
-default startupitem.start       ""
-default startupitem.stop        ""
+default startupitem.requires    {}
+default startupitem.restart     {}
+default startupitem.start       {}
+default startupitem.stop        {}
 default startupitem.type        {[portstartupitem::get_startupitem_type]}
 default startupitem.uniquename  {org.macports.${startupitem.name}}
+default startupitem.user        {}
 
 default startupitem.daemondo.verbosity  1
 
@@ -132,44 +151,45 @@ proc portstartupitem::get_startupitem_type {} {
 # each startupitem that has been defined in the portfile
 proc portstartupitem::foreach_startupitem {body} {
     global startupitems
-    set vars [list autostart debug create executable init install location \
-              logevents logfile name netchange pidfile plist requires \
-              restart start stop type uniquename daemondo.verbosity]
+    set vars [list autostart create custom_file debug executable group \
+              init install location logevents logfile logfile.stderr \
+              name netchange pidfile plist requires restart start stop type \
+              uniquename user daemondo.verbosity]
 
-    array set startupitems_dict {}
+    set startupitems_dict [dict create]
     if {[info exists startupitems] && $startupitems ne ""} {
         foreach {key val} $startupitems {
             if {$key eq "name"} {
                 set curname $val
                 # these have defaults based on the name
                 set uniquename org.macports.${val}
-                lappend startupitems_dict($curname) uniquename $uniquename
-                lappend startupitems_dict($curname) plist ${uniquename}.plist
+                dict set startupitems_dict $curname uniquename $uniquename
+                dict set startupitems_dict $curname plist ${uniquename}.plist
             }
-            lappend startupitems_dict($curname) $key [list $val]
+            dict set startupitems_dict $curname $key $val
         }
     } else {
         global startupitem.name
         foreach var $vars {
             global startupitem.${var}
             if {[info exists startupitem.${var}]} {
-                lappend startupitems_dict(${startupitem.name}) $var [set startupitem.${var}]
+                dict set startupitems_dict ${startupitem.name} $var [set startupitem.${var}]
             }
         }
     }
 
-    uplevel 1 "set si_vars [list $vars]"
-    foreach item [array names startupitems_dict] {
-        uplevel 1 "array unset si_dict; array set si_dict [list $startupitems_dict($item)]"
+    uplevel 1 [list set si_vars $vars]
+    dict for {item subdict} $startupitems_dict {
+        uplevel 1 [list set si_dict $subdict]
         uplevel 1 {
             foreach si_var $si_vars {
-                if {[info exists si_dict($si_var)]} {
-                    set si_${si_var} $si_dict($si_var)
+                if {[dict exists $si_dict $si_var]} {
+                    set si_${si_var} [dict get $si_dict $si_var]
                 } else {
                     global startupitem.${si_var}
                     if {[info exists startupitem.${si_var}]} {
                         set si_${si_var} [set startupitem.${si_var}]
-                        set si_dict($si_var) [set startupitem.${si_var}]
+                        dict set si_dict $si_var [set startupitem.${si_var}]
                     }
                 }
             }
@@ -181,8 +201,8 @@ proc portstartupitem::foreach_startupitem {body} {
 # Add user notes regarding any installed startupitem
 proc portstartupitem::add_notes {} {
     global subport startupitem_autostart
-    set autostart_names {}
-    set normal_names {}
+    set autostart_names [list]
+    set normal_names [list]
 
     foreach_startupitem {
         if {$si_type eq "none"} {
@@ -249,37 +269,56 @@ proc portstartupitem::add_notes {} {
 # Register the above procedure as a callback after Portfile evaluation
 port::register_callback portstartupitem::add_notes
 
+# Helper: link given .plist into the correct location
+proc portstartupitem::install_darwin_launchd {srcpath dstdir install} {
+    global destroot prefix
+    if {[getuid] == 0 && $install} {
+        file mkdir ${destroot}/Library/${dstdir}
+        ln -sf $srcpath ${destroot}/Library/${dstdir}
+    } else {
+        ln -sf $srcpath ${destroot}${prefix}/etc/${dstdir}
+    }
+}
+
 proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
     global UI_PREFIX prefix destroot destroot.keepdirs subport macosx_deployment_target
 
-    array set si $attrs
-
-    set scriptdir ${prefix}/etc/startup
-
-    set itemname        $si(name)
-    set uniquename      $si(uniquename)
-    set plistname       $si(plist)
-    set daemondest      $si(location)
+    set uniquename      [dict get $attrs uniquename]
+    set plistname       [dict get $attrs plist]
+    set daemondest      [dict get $attrs location]
     set itemdir         ${prefix}/etc/${daemondest}/${uniquename}
-    set args            [list \
-                          "${prefix}/bin/daemondo" \
-                          "--label=${itemname}" \
-                        ]
 
     file mkdir ${destroot}${itemdir}
     if {[getuid] == 0} {
         file attributes ${destroot}${itemdir} -owner root -group wheel
     }
 
-    if {$si(executable) ne "" &&
-        $si(init) eq "" &&
-        $si(start) eq "" &&
-        $si(stop) eq "" &&
-        $si(restart) eq ""} {
+    if {[dict get $attrs custom_file] ne ""} {
+        # The port is supplying its own plist
+        file copy [dict get $attrs custom_file] ${destroot}${itemdir}/${plistname}
+        install_darwin_launchd ${itemdir}/${plistname} $daemondest [dict get $attrs install]
+        return
+    }
+
+    set scriptdir ${prefix}/etc/startup
+
+    set itemname        [dict get $attrs name]
+    set username        [dict get $attrs user]
+    set groupname       [dict get $attrs group]
+    set args            [list \
+                          "${prefix}/bin/daemondo" \
+                          "--label=${itemname}" \
+                        ]
+
+    if {[dict get $attrs executable] ne "" &&
+        [dict get $attrs init] eq "" &&
+        [dict get $attrs start] eq "" &&
+        [dict get $attrs stop] eq "" &&
+        [dict get $attrs restart] eq ""} {
 
         # An executable is specified, and there is no init, start, stop, or restart
         # code; so we don't need a wrapper script
-        set args [concat $args "--start-cmd" $si(executable) ";"]
+        set args [concat $args "--start-cmd" [dict get $attrs executable] ";"]
 
     } else {
 
@@ -289,14 +328,14 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
         set wrappername     ${itemname}.wrapper
         set wrapper         "${itemdir}/${wrappername}"
 
-        if {$si(start) eq ""} {
-            set si(start) [list "sh ${scriptdir}/${subport}.sh start"]
+        if {[dict get $attrs start] eq ""} {
+            dict set attrs start [list "sh ${scriptdir}/${subport}.sh start"]
         }
-        if {$si(stop) eq ""} {
-            set si(stop) [list "sh ${scriptdir}/${subport}.sh stop"]
+        if {[dict get $attrs stop] eq ""} {
+            dict set attrs stop [list "sh ${scriptdir}/${subport}.sh stop"]
         }
-        if {$si(restart) eq ""} {
-            set si(restart) [list Stop Start]
+        if {[dict get $attrs restart] eq ""} {
+            dict set attrs restart [list Stop Start]
         }
 
         lappend args \
@@ -320,7 +359,7 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
         puts ${item} "# Init"
         puts ${item} "#"
         puts ${item} "prefix=$prefix"
-        foreach line $si(init)    { puts ${item} ${line} }
+        foreach line [dict get $attrs init]    { puts ${item} ${line} }
         puts ${item} ""
 
         puts ${item} "#"
@@ -328,7 +367,7 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
         puts ${item} "#"
         puts ${item} "Start()"
         puts ${item} "\{"
-        foreach line $si(start)   { puts ${item} "\t${line}" }
+        foreach line [dict get $attrs start]   { puts ${item} "\t${line}" }
         puts ${item} "\}"
         puts ${item} ""
 
@@ -337,7 +376,7 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
         puts ${item} "#"
         puts ${item} "Stop()"
         puts ${item} "\{"
-        foreach line $si(stop)    { puts ${item} "\t${line}" }
+        foreach line [dict get $attrs stop]    { puts ${item} "\t${line}" }
         puts ${item} "\}"
         puts ${item} ""
 
@@ -346,7 +385,7 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
         puts ${item} "#"
         puts ${item} "Restart()"
         puts ${item} "\{"
-        foreach line $si(restart) { puts ${item} "\t${line}" }
+        foreach line [dict get $attrs restart] { puts ${item} "\t${line}" }
         puts ${item} "\}"
         puts ${item} ""
 
@@ -373,12 +412,12 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
         close ${item}
     }
 
-    if {$si(netchange)} {
+    if {[dict get $attrs netchange]} {
         lappend args "--restart-netchange"
     }
 
     # To log events then tell daemondo to log at verbosity=n
-    if {$si(logevents)} {
+    if {[dict get $attrs logevents]} {
         lappend args "--verbosity=[option startupitem.daemondo.verbosity]"
     }
 
@@ -390,7 +429,7 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
     #   (3) clean [pidfilename]
     #   (4) manual [pidfilename]
     #
-    set pidfileArgCnt [llength $si(pidfile)]
+    set pidfileArgCnt [llength [dict get $attrs pidfile]]
     if {${pidfileArgCnt} > 0} {
         if { $pidfileArgCnt == 1 } {
             set pidFile "${prefix}/var/run/${itemname}.pid"
@@ -398,7 +437,7 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
                 lappend destroot.keepdirs "${destroot}${prefix}/var/run"
             }
         } else {
-            set pidFile [lindex $si(pidfile) 1]
+            set pidFile [lindex [dict get $attrs pidfile] 1]
         }
 
         if {${pidfileArgCnt} > 2} {
@@ -406,7 +445,7 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
         }
 
         # Translate into appropriate arguments to daemondo
-        set pidStyle [lindex $si(pidfile) 0]
+        set pidStyle [lindex [dict get $attrs pidfile] 0]
         switch -- ${pidStyle} {
             none    { lappend args "--pid=none" }
             auto    { lappend args "--pid=fileauto" "--pidfile" ${pidFile} }
@@ -417,7 +456,7 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
             }
         }
     } else {
-        if {$si(executable) ne ""} {
+        if {[dict get $attrs executable] ne ""} {
             lappend args "--pid=exec"
         } else {
             lappend args "--pid=none"
@@ -441,17 +480,25 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
     puts ${plist} "</array>"
 
     puts ${plist} "<key>Disabled</key><true/>"
-    if {$macosx_deployment_target ne "10.4"} {
-        puts ${plist} "<key>KeepAlive</key><true/>"
-    } else {
-        puts ${plist} "<key>OnDemand</key><false/>"
+    puts ${plist} "<key>KeepAlive</key><true/>"
+
+    if {$username ne ""} {
+        puts ${plist} "<key>UserName</key><string>$username</string>"
     }
 
-    if {$si(logfile) ne ""} {
-        puts ${plist} "<key>StandardOutPath</key><string>$si(logfile)</string>"
+    if {$groupname ne ""} {
+        puts ${plist} "<key>GroupName</key><string>$groupname</string>"
     }
 
-    if {$si(debug)} {
+    if {[dict get $attrs logfile] ne ""} {
+        puts ${plist} "<key>StandardOutPath</key><string>[dict get $attrs logfile]</string>"
+    }
+
+    if {[dict get $attrs logfile.stderr] ne ""} {
+        puts ${plist} "<key>StandardErrorPath</key><string>[dict get $attrs logfile.stderr]</string>"
+    }
+
+    if {[dict get $attrs debug]} {
         puts ${plist} "<key>Debug</key><true/>"
     }
 
@@ -460,23 +507,22 @@ proc portstartupitem::startupitem_create_darwin_launchd {attrs} {
 
     close ${plist}
 
-    if {[getuid] == 0 && $si(install)} {
-        file mkdir "${destroot}/Library/${daemondest}"
-        ln -sf "${itemdir}/${plistname}" "${destroot}/Library/${daemondest}"
-    } else {
-        ln -sf ${itemdir}/${plistname} ${destroot}${prefix}/etc/${daemondest}
-    }
+    install_darwin_launchd ${itemdir}/${plistname} $daemondest [dict get $attrs install]
 }
 
 proc portstartupitem::startupitem_create {} {
     global UI_PREFIX
 
     foreach_startupitem {
-        if {${si_type} ne "none" && [tbool si_create]} {
-            ui_notice "$UI_PREFIX [msgcat::mc "Creating ${si_type} control script '$si_name'"]"
+        if {${si_type} ne "none" && ([tbool si_create] || $si_custom_file ne "")} {
+            if {[tbool si_create]} {
+                ui_debug "Creating ${si_type} control script '$si_name'"
+            } else {
+                ui_debug "Installing ${si_type} control script '$si_name'"
+            }
 
             switch -- ${si_type} {
-                launchd         { startupitem_create_darwin_launchd [array get si_dict] }
+                launchd         { startupitem_create_darwin_launchd $si_dict }
                 default         { ui_error "$UI_PREFIX [msgcat::mc "Unrecognized startupitem type %s" ${si_type}]" }
             }
         }
@@ -489,9 +535,9 @@ proc portstartupitem::loaded {} {
     set launchctl_path ${portutil::autoconf::launchctl_path}
     if {$launchctl_path eq ""} {
         # assuming not loaded if there's no launchctl
-        return {}
+        return [list]
     }
-    set ret {}
+    set ret [list]
     global os.major sudo_user
     foreach_startupitem {
         if {$si_type ne "launchd"} {
@@ -513,12 +559,8 @@ proc portstartupitem::loaded {} {
             if {![catch {exec -ignorestderr $launchctl_path print ${domain}/${si_uniquename} >&/dev/null}]} {
                 lappend ret $si_name
             }
-        } elseif {${os.major} >= 9} {
-            if {![catch {exec_as_uid $uid {system "$launchctl_path list ${si_uniquename} > /dev/null"}}]} {
-                lappend ret $si_name
-            }
         } else {
-            if {![catch {exec_as_uid $uid {system "$launchctl_path list | grep -F ${si_uniquename} > /dev/null"}}]} {
+            if {![catch {exec_as_uid $uid {system "$launchctl_path list ${si_uniquename} > /dev/null"}}]} {
                 lappend ret $si_name
             }
         }
